;; Generated package description from denote-menu.el  -*- no-byte-compile: t -*-
(define-package "denote-menu" "1.3.0" "View denote files in a tabulated list." '((emacs "28.1") (denote "3.0.0")) :commit "d3b890648404eb4df2dabbb3d783105672e0ea23" :authors '(("Mohamed Suliman" . "sulimanm@tcd.ie")) :maintainer '("Mohamed Suliman" . "sulimanm@tcd.ie") :url "https://github.com/namilus/denote-menu")
